import os
import subprocess
from typing import Dict, List # pylint: disable=unused-import

from sheepdoge.exception import SheepdogeShellRunnerException


class ShellRunner(object):
    """A wrapper around the `subprocess.check_call` command, to DRY the
    transformations necessary to run a shell command.

    :param cmd_as_list: The command for the `ShellRunner` to execute.
    :raises: SheepdogeShellRunnerException
    """
    def __init__(self, cmd_as_list):
        # type: (List[str]) -> None
        self._cmd_as_str = ' '.join(cmd_as_list)

    def run(self, env_additions=None):
        # type: (Dict[str, str]) -> None
        """Execute the command wrapped by `ShellRunner`.

        :param env_additions: Optional method for specifying additional
        environment variables for the shell in the subprocess in which the
        command will execute.
        """
        env_additions = env_additions or {}

        shell_env = os.environ.copy()
        shell_env.update(env_additions)

        try:
            subprocess.check_call(
                self._cmd_as_str,
                shell=True,
                env=shell_env
            )
        except subprocess.CalledProcessError as err:
            raise SheepdogeShellRunnerException(
                '{} failed: {}'.format(self._cmd_as_str, str(err))
            )



from configparser import ConfigParser, NoOptionError
from typing import Dict # pylint: disable=unused-import
import os

from sheepdoge.exception import (
    SheepdogeConfigurationAlreadyInitializedException,
    SheepdogeConfigurationNotInitializedException)


DEFAULTS = {
    'kennel_playbook_path': 'kennel.yml',
    'kennel_roles_path': '.kennel_roles',
    'pupfile_path': 'pupfile.yml',
    'vault_password_file': None
} # type: Dict[str, str]


class Config(object):
    """Config class for which there should only be one instance at anytime.
    Additionally, we can only set the config values during initialization.
    Multiple different classes can access this single instance at a time.
    """
    _config = None # type: Config


    def __init__(self, config_dict):
        # type: (Dict[str, str]) -> None
        self._config_dict = config_dict

    @classmethod
    def clear_config_singleton(cls):
        # type: () -> None
        """Delete the current configuration singleton to allow the
        initialization of a new one. This method is predominantly used
        during test.
        """
        cls._config = None

    @classmethod
    def get_config_singleton(cls):
        # type: () -> Config
        """Return the current config singleton instance. We must initialize
        the singleton before calling this method.

        :return: The singleton instance.
        """
        if cls._config is None:
            raise SheepdogeConfigurationNotInitializedException
        return cls._config

    @classmethod
    def initialize_config_singleton(cls, config_file_contents=None,
                                    config_options=None):
        # type: (str, Dict[str, str]) -> None
        """Initialize the config singleton with the proper values. If we
        specify no additional values during configuration, then the config
        will contain all defaults. We can, in priority order, pass in the
        contents of a *.cfg file and a dictionary of options. Typically we
        derive this dictionary of options from the command line.

        Finally, after setting all of the base configuration values,
        we compute additional configuration values which are useful
        throughout the program.

        :param config_file_contents: The str contents of the .cfg file
        containing kennel configuration.
        :param config_options: The dict specifying the highest priority
        configuration values.
        """
        if cls._config is not None:
            raise SheepdogeConfigurationAlreadyInitializedException()

        config_dict = {} # type: Dict[str, str]

        cls._set_config_default_values(config_dict)

        if config_file_contents:
            cls._set_config_file_values(config_dict, config_file_contents)

        if config_options:
            cls._set_config_option_values(config_dict, config_options)

        cls._set_calculated_config_values(config_dict)

        cls._config = cls(config_dict)

    @classmethod
    def _set_config_default_values(cls, config_dict):
        # type: (Dict[str, str]) -> None
        """Set defaults for all views here - they will be overwritten in the
        following steps if necessary.
        """
        config_dict.update(DEFAULTS)

    @classmethod
    def _set_config_file_values(cls, config_dict, config_file_contents):
        # type: (Dict[str, str], str) -> None
        config_parser = ConfigParser()
        config_parser.read_string(config_file_contents)

        kennel_cfg_section = 'kennel'

        for currently_defined_key in config_dict.keys():
            try:
                config_file_value = config_parser.get(kennel_cfg_section,
                                                      currently_defined_key)

                config_dict[currently_defined_key] = config_file_value
            except NoOptionError:
                pass  # If the value isn't specified, skip

    @classmethod
    def _set_config_option_values(cls, config_dict, config_options):
        # type: (Dict[str, str], Dict[str, str]) -> None
        config_dict.update(config_options)

    @classmethod
    def _set_calculated_config_values(cls, config_dict):
        # type: (Dict[str, str]) -> None
        pupfile_path = config_dict['pupfile_path']
        pupfile_dir = os.path.dirname(os.path.realpath(pupfile_path))

        kennel_roles_path = config_dict['kennel_roles_path']
        abs_kennel_roles_dir = os.path.realpath(kennel_roles_path)

        calculated_config = {
            'abs_pupfile_dir': pupfile_dir,
            'abs_kennel_roles_dir': abs_kennel_roles_dir
        }

        config_dict.update(calculated_config)

    def get(self, key):
        # type: (str) -> str
        """Retrieve the value for the given configuration key.

        :param key: One of the available configuration options.
        """
        return self._config_dict[key]



"""Top-level module for sheepdoge actions"""

from sheepdoge.config import Config


class Action(object):
    """A base class for all actions.

    :param config: The configuration object for performing this action.
    """
    def __init__(self, config=None):
        # type: (Config) -> None
        self._config = config or Config.get_config_singleton()

    def run(self):
        # type: () -> None
        """Execute this action."""
        self._setup()
        self._execute()
        self._teardown()

    def _setup(self):
        # type: () -> None
        pass

    def _execute(self):
        # type: () -> None
        pass

    def _teardown(self):
        # type: () -> None
        pass



class SheepdogeException(Exception):
    pass


class SheepdogeConfigurationAlreadyInitializedException(SheepdogeException):
    pass


class SheepdogeConfigurationNotInitializedException(SheepdogeException):
    pass


class SheepdogeInvalidPupTypeException(SheepdogeException):
    pass


class SheepdogeShellRunnerException(SheepdogeException):
    pass



"""Module for actions related to running a kennel."""

from typing import Any # pylint: disable=unused-import

from sheepdoge.action import Action
from sheepdoge.kennel import Kennel # pylint: disable=unused-import

class RunAction(Action):
    def __init__(self, kennel, *args, **kwargs):
        # type: (Kennel, *Any, **Any) -> None
        super(RunAction, self).__init__(*args, **kwargs)

        self._kennel = kennel

    def _execute(self):
        # type: () -> None
        self._kennel.run()



"""Module for code relating to installing pups for a kennel."""

from concurrent.futures import ThreadPoolExecutor, wait
import multiprocessing
from typing import Any, Iterable, Type # pylint: disable=unused-import

from sheepdoge.action import Action
from sheepdoge.kennel import Kennel # pylint: disable=unused-import
from sheepdoge.pup import Pup # pylint: disable=unused-import


class InstallAction(Action):
    def __init__(self, kennel_cls, pup_cls, *args, **kwargs):
        # type: (Type[Kennel], Type[Pup], *Any, **Any) -> None
        super(InstallAction, self).__init__(*args, **kwargs)

        self._pups_to_install = [] # type: Iterable[Pup]

        self._kennel_cls = kennel_cls
        self._pup_cls = pup_cls

    def _setup(self):
        # type: () -> None
        self._kennel_cls.refresh_roles(
            self._config.get('abs_kennel_roles_dir'))
        self._pups_to_install = self._pup_cls.parse_pupfile_into_pups(
            self._config.get('pupfile_path'))

    def _execute(self):
        # type: () -> None
        for pup in self._pups_to_install:
            pup.install()


class ParallelInstallAction(InstallAction):
    """An InstallAction which downloads pups in parallel."""
    def __init__(self, kennel_cls, pup_cls, max_workers=None, **kwargs):
        # type: (Type[Kennel], Type[Pup], int, **Any) -> None
        super(ParallelInstallAction, self).__init__(
            kennel_cls, pup_cls, **kwargs)

        self._max_workers = (max_workers or
                             self._get_default_max_workers_from_cores())

    @staticmethod
    def _get_default_max_workers_from_cores():
        # type: () -> int
        return multiprocessing.cpu_count()

    def _execute(self):
        # type: () -> None
        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            install_futures = {
                executor.submit(pup.install) for pup in self._pups_to_install
            }

            wait(install_futures)



"""Configuration for the `sheepdoge` pip package."""

import ast
import re

from setuptools import setup, find_packages

GITHUB_URL = 'https://github.com/sheepdoge/sheepdoge'

with open('sheepdoge/__version__.py', 'rb') as version_file:
    file_contents = version_file.read().decode('utf-8')
    version_re = re.compile(r'__version__\s+=\s+(.*)')
    version = str(ast.literal_eval(version_re.search(file_contents).group(1)))

with open('requirements.in', 'r') as requirements_file:
    install_requires = [package.strip() for package in requirements_file.readlines()]

setup(
    name='sheepdoge',
    version=version,
    description='Manage your personal Unix machine(s) with Ansible.',
    url=GITHUB_URL,
    author='Matt McNaughton',
    author_email='mattjmcnaughton@gmail.com',
    license='Apache',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    keywords=['provisioning', 'automation'],
    packages=find_packages(where='.', exclude=['tests*']),
    install_requires=install_requires,
    entry_points={
        'console_scripts': ['sheepdoge = sheepdoge.cli:main']
    }
)



from collections import namedtuple
from contextlib import contextmanager
import os
import shutil
import tempfile
from typing import Any, Dict, Iterable, Iterator, List, Tuple, Type # pylint: disable=unused-import

import yaml

from sheepdoge.config import Config
from sheepdoge.exception import SheepdogeInvalidPupTypeException
from sheepdoge.utils import ShellRunner

# The character used in the `location` in the pupfile to split between
# `pup_type` and `path`.
LOCATION_SPLIT_CHAR = '+'

PupfileEntry = namedtuple('PupfileEntry', 'name path pup_type')


def _pup_types_to_classes():
    # type: () -> Dict[str, Type[Pup]]
    return {
        'fs': FsPup,
        'galaxy': GalaxyPup,
        'git': GitPup
    }


def _dependency_types_to_classes():
    # type: () -> Dict[str, Type[PupDependencies]]
    return {
        'txt': PythonDependencies,
        'yml': AnsibleDependencies
    }


class Pup(object):
    """Container of all pup related logic. Contains both static methods as well
    as base methods inherited by the different types of pups.
    """

    PYTHON_DEP_FILE = 'requirements.txt'
    ANSIBLE_DEP_FILE = 'requirements.yml'


    def __init__(self, name, path, config=None):
        # type: (str, str, Config) -> None
        self._name = name
        self._path = path
        self._pup_dependencies = [] # type: List[PupDependencies]

        self._config = config or Config.get_config_singleton()

    @classmethod
    def parse_pupfile_into_pups(cls, pupfile_path):
        # type: (str) -> List[Pup]
        """Return a list of `Pup` objects for each pup we wish to install."""
        # Read in file and parse with yml
        with open(pupfile_path, 'r') as pupfile:
            entries_from_file = cls.parse_text_into_entries(pupfile.read())

        return cls.create_from_entries(entries_from_file)

    @classmethod
    def parse_text_into_entries(cls, file_contents):
        # type: (str) -> List[PupfileEntry]
        """Given the text of `pupfile.yml`, return the structured data we will
        iterate through to create individual `Pup` instances.

        :param file_contents: A string containing the contents of `pupfile.yml`.
        :return: A list of structured pupfile entries.
        """
        entries = []

        for dict_entry in yaml.load(file_contents):
            name = dict_entry['name']
            pup_type, path = cls._parse_location(dict_entry['location'])

            if pup_type not in _pup_types_to_classes().keys():
                err_msg = '{} is not a valid pup type.'.format(pup_type)
                raise SheepdogeInvalidPupTypeException(err_msg)
            entries.append(PupfileEntry(name=name, path=path, pup_type=pup_type))

        return entries

    @staticmethod
    def _parse_location(location):
        # type: (str) -> Tuple[str, str]
        split_location = location.split(LOCATION_SPLIT_CHAR)

        return split_location[0], split_location[1]

    @classmethod
    def create_from_entries(cls, entries):
        # type: (Iterable[PupfileEntry]) -> List[Pup]
        """Create a pup based on a line in a `pupfile.yml`.

        Instantiates and returns a subclass of `Pup` for the specific type of
        pup we are installing (i.e. from local filesystem, git, ansible-galaxy).

        :param entries: The input entries from the parsed pupfile.
        :return: The specific pup instances we are installing.
        """
        pups = []

        for entry in entries:
            pup_cls = _pup_types_to_classes()[entry.pup_type]
            pup = pup_cls(entry.name, entry.path)
            pups.append(pup)

        return pups

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Return a readable version of the pup"""
        # @TODO(mattjmcnaughton) Determine the best long term solution. Is it
        # defining `__repr__` or maybe `__eq__`?
        return {
            'name': self._name,
            'pup_type': self.__class__,
            'path': self._path
        }

    def install(self):
        # type: () -> None
        """Install all aspects of the pup.
        """
        pup_in_kennel_path = self._install_pup()
        self._install_pup_dependencies(pup_in_kennel_path)

    def _install_pup(self):
        # type: () -> str
        """Download the pup onto the local file system."""
        raise NotImplementedError

    def _install_pup_dependencies(self, pup_in_kennel_path):
        # type: (str) -> None
        """Install the dependencies for the pup (i.e. other ansible roles or
        python packages.
        """
        python_dependency_file, ansible_dependency_file = self._get_dep_files(
            pup_in_kennel_path
        )

        for dep_file_path in {python_dependency_file, ansible_dependency_file}:
            if dep_file_path is not None:
                dependencies = PupDependencies.create_from_dep_file_path(dep_file_path)
                dependencies.install()

    @classmethod
    def _get_dep_files(cls, pup_in_kennel_path):
        # type: (str) -> Tuple[str, str]
        python_dep_file = os.path.join(pup_in_kennel_path, cls.PYTHON_DEP_FILE)
        ansible_dep_file = os.path.join(pup_in_kennel_path, cls.ANSIBLE_DEP_FILE)

        if not os.path.isfile(python_dep_file):
            python_dep_file = None

        if not os.path.isfile(ansible_dep_file):
            ansible_dep_file = None

        return python_dep_file, ansible_dep_file

    @staticmethod
    @contextmanager
    def _with_tmp_dir():
        # type: () -> Iterator[str]
        tmp_dir_path = tempfile.mkdtemp()

        yield tmp_dir_path

        shutil.rmtree(tmp_dir_path)


class FsPup(Pup):
    """A pup for which the source code is already on the local file system (we
    predominantly use this pup for testing).

    All fs pupfile locations should be specified relative to the location of the
    pupfile.
    """
    def _install_pup(self):
        # type: () -> str
        fs_pup_location = os.path.join(self._config.get('abs_pupfile_dir'),
                                       self._path)

        pup_in_kennel_path = os.path.join(
            self._config.get('abs_kennel_roles_dir'), self._name)

        shutil.copytree(fs_pup_location, pup_in_kennel_path)

        return pup_in_kennel_path


class GitPup(Pup):
    """A pup for which the source lives in a remote git repo.
    """
    def _install_pup(self):
        # type: () -> str
        with self._with_tmp_dir() as install_dir:
            install_path = os.path.join(install_dir, self._name)

            git_cmd = [
                'git',
                'clone',
                self._path,
                install_path
            ]

            ShellRunner(git_cmd).run()

            pup_in_kennel_path = os.path.join(
                self._config.get('abs_kennel_roles_dir'),
                self._name
            )

            shutil.copytree(install_path, pup_in_kennel_path)

            return pup_in_kennel_path


class GalaxyPup(Pup):
    """A pup for which the source lives on ansible-galaxy.
    """
    def _install_pup(self):
        # type: () -> str
        with self._with_tmp_dir() as install_dir:
            ansible_galaxy_cmd = [
                'ansible-galaxy',
                'install',
                self._path,
                '-p',
                install_dir
            ]

            ShellRunner(ansible_galaxy_cmd).run()

            role_name = os.listdir(install_dir)[0]

            install_path = os.path.join(
                install_dir,
                role_name
            )

            pup_in_kennel_path = os.path.join(
                self._config.get('abs_kennel_roles_dir'),
                self._name
            )

            shutil.copytree(install_path, pup_in_kennel_path)

            return pup_in_kennel_path


class PupDependencies(object):
    """A base representation of a collection of pup dependencies. Either
    python packages or ansible roles.
    """
    @staticmethod
    def create_from_dep_file_path(dep_file_path):
        # type: (str) -> PupDependencies
        """Given the path to a pup's dependency file, instantiate an instance of
        the appropriate subclass based on the given `dep_file_path`.

        :param dep_file_path: The path to the dependency file.
        :return: Instance of subclass of `PupDependencies`.
        """
        file_extension = dep_file_path.split('.')[-1]

        return _dependency_types_to_classes()[file_extension](dep_file_path)

    def __init__(self, dep_file_path, config=None):
        # type: (str, Config) -> None
        self._dep_file_path = dep_file_path

        self._config = config or Config.get_config_singleton()

    def install(self):
        # type: () -> None
        """Install the pup dependency."""
        raise NotImplementedError


class PythonDependencies(PupDependencies):
    """Pip packages that must be installed on local machine for this kennel to
    run.
    """
    def install(self):
        # type: () -> None
        pip_cmd = [
            'pip',
            'install',
            '-r',
            self._dep_file_path
        ]

        ShellRunner(pip_cmd).run()


class AnsibleDependencies(PupDependencies):
    """Ansible roles that must be installed on local machine for this kennel
    to run.
    """
    def install(self):
        # type: () -> None
        ansible_galaxy_cmd = [
            'ansible-galaxy',
            'install',
            '-r',
            self._dep_file_path,
            '-p',
            self._config.get('abs_kennel_roles_dir')
        ]

        ShellRunner(ansible_galaxy_cmd).run()



import os
import shutil
from typing import List # pylint: disable=unused-import

from sheepdoge.config import Config
from sheepdoge.utils import ShellRunner


class Kennel(object):
    def __init__(self, additional_ansible_args='', config=None):
        # type: (str, Config) -> None
        self._additional_ansible_args = additional_ansible_args
        self._config = config or Config.get_config_singleton()

    @staticmethod
    def refresh_roles(kennel_roles_path):
        # type: (str) -> None
        """Ensure a clean directory exists at `kennel_roles_path`."""
        if os.path.isdir(kennel_roles_path) and os.listdir(kennel_roles_path):
            shutil.rmtree(kennel_roles_path)
            os.mkdir(kennel_roles_path)

    def run(self):
        # type: () -> None
        ansible_playbook_cmd = [
            'ansible-playbook',
            self._config.get('kennel_playbook_path')
        ]

        self._append_vault_password_file_flag_if_file_exists(
            ansible_playbook_cmd)

        if self._additional_ansible_args:
            ansible_playbook_cmd += self._additional_ansible_args.split(' ')

        ShellRunner(ansible_playbook_cmd).run(env_additions={
            'ANSIBLE_ROLES_PATH': self._config.get('kennel_roles_path')
        })

    def _append_vault_password_file_flag_if_file_exists(self,
                                                        ansible_playbook_cmd):
        # type: (List[str]) -> None
        password_file = self._config.get('vault_password_file')

        if password_file:
            ansible_playbook_cmd.append(
                '--vault-password-file={}'.format(password_file)
            )



"""The point of entry for the sheepdoge pip package console script"""

import io
from typing import Dict # pylint: disable=unused-import

import click

from sheepdoge.action.install import InstallAction, ParallelInstallAction
from sheepdoge.action.run import RunAction
from sheepdoge.app import Sheepdoge
from sheepdoge.config import Config
from sheepdoge.pup import Pup
from sheepdoge.kennel import Kennel
from sheepdoge.__version__ import __version__


def _initialize_config(config_file, config_options=None):
    # type: (str, Dict[str, str]) -> None
    config_options = config_options or {}

    with io.open(config_file, 'r',
                 encoding='utf-8') as config_file_open_for_reading:
        config_file_contents = config_file_open_for_reading.read()

    Config.initialize_config_singleton(
        config_file_contents=config_file_contents,
        config_options=config_options
    )


@click.group()
@click.version_option(prog_name='sheepdoge', version=__version__)
def cli():
    # type: () -> None
    pass


@cli.command()
@click.option('--config-file', default='kennel.cfg')
@click.option('--parallel/--no-parallel', default=True)
def install(config_file, parallel):
    # type: (str, bool) -> None
    _initialize_config(config_file)

    install_cls = ParallelInstallAction if parallel else InstallAction

    install_action = install_cls(Kennel, Pup)
    Sheepdoge(install_action).run()


@cli.command()
@click.option('--ansible-args', type=str)
@click.option('--config-file', default='kennel.cfg')
def run(ansible_args, config_file):
    # type: (str, str) -> None
    _initialize_config(config_file)

    kennel = Kennel(additional_ansible_args=ansible_args)
    run_action = RunAction(kennel)
    Sheepdoge(run_action).run()


def main():
    # type: () -> None
    cli()



"""Orchestrates the different `sheepdoge` operations."""

from sheepdoge.config import Config
from sheepdoge.action import Action # pylint: disable=unused-import


class Sheepdoge(object):
    """A class we instantiate with instances of the `Action`, which indicate
    which cli command we'll perform.

    :param action: The Sheepdoge action we're running.
    """
    def __init__(self, action, config=None):
        # type: (Action, Config) -> None
        self._action = action
        self._config = config or Config.get_config_singleton()

    def run(self):
        # type: () -> None
        """Execute an command given to `sheepdoge`."""
        self._action.run()
